//! Provides [Atomic]-[Mentionable] interface.

use std::ops::Deref;

use super::*;

/// Generic implementation of a [Mentionable] for [Atomic]s.
#[derive(Clone)]
pub struct AtomicObject<A: AtomicBase> {
    atomic: A,
}

impl<A: AtomicBase> From<A> for AtomicObject<A> {
    fn from(value: A) -> Self {
        Self { atomic: value }
    }
}

impl<A: AtomicBase> AsRef<A> for AtomicObject<A> {
    fn as_ref(&self) -> &A {
        &self.atomic
    }
}

impl<A: AtomicBase> Deref for AtomicObject<A> {
    type Target = A;

    fn deref(&self) -> &Self::Target {
        &self.atomic
    }
}

impl<A: AtomicBase> Serializable for AtomicObject<A> {
    fn serialize(&self, serializer: &mut dyn Serializer) {
        self.atomic.serialize(serializer)
    }
}

pub trait AoProxy<'a, Ctx: Context<'a>>: AtomicProxy {
    type Mtbl: MentionableBase<'a, Ctx, Fctr = Self::Fctr> + MentionableTop<'a, Ctx>;
    type Fctr: FactoryBase<'a, Ctx, Mtbl = Self::Mtbl>;

    fn factory() -> Self::Fctr;
    fn wrap(a: Self::A) -> Self::Mtbl;
}

impl<'a, Ctx: Context<'a>, A: AtomicBase + WithParseMode> MentionableBase<'a, Ctx>
    for AtomicObject<A>
where
    A::WithMode: AoProxy<'a, Ctx, Mtbl = Self, A = A>,
{
    type Fctr = <A::WithMode as AoProxy<'a, Ctx>>::Fctr;

    fn factory(&self) -> Self::Fctr {
        <A::WithMode as AoProxy<'a, Ctx>>::factory()
    }
}

impl<'a, Ctx: Context<'a>, A: AtomicBase> MentionableTop<'a, Ctx> for AtomicObject<A> {
    fn topology(&self) -> Hash {
        Ctx::hash(b"")
    }

    fn points_typed(&self, _points: &mut impl PointsVisitor<'a, Ctx>) {}
}

/// Generic implementation of a [Factory] for [Atomic]s.
pub struct AtomicFactory<A> {
    _pd: PhantomData<A>,
}

impl<A> AtomicFactory<A> {
    fn new() -> Self {
        AtomicFactory { _pd: PhantomData }
    }
}

impl<A: AtomicBase> Clone for AtomicFactory<A> {
    fn clone(&self) -> Self {
        Self::new()
    }
}

impl<'a, Ctx: Context<'a>, A: AtomicBase + WithParseMode> FactoryBase<'a, Ctx> for AtomicFactory<A>
where
    A::WithMode: AoProxy<'a, Ctx, Fctr = Self, A = A>,
{
    type Mtbl = <A::WithMode as AoProxy<'a, Ctx>>::Mtbl;

    type ParseError = A::AParseError;
}

impl<A: ParseMode> ParseMode for AtomicFactory<A> {
    type Mode = A::Mode;
}

impl<'a, Ctx: Context<'a>, A: RegularAtomic> AoProxy<'a, Ctx> for WithMode<A, RegularMode> {
    type Mtbl = AtomicObject<A>;

    type Fctr = AtomicFactory<A>;

    fn factory() -> Self::Fctr {
        AtomicFactory::new()
    }

    fn wrap(a: Self::A) -> Self::Mtbl {
        a.into()
    }
}

impl<'a, Ctx: Context<'a>, A: RegularAtomic> RegularFactory<'a, Ctx> for AtomicFactory<A> {
    fn rdeserialize(&self, inctx: impl InCtx<'a, Ctx>) -> ParseResult<'a, Ctx, Self> {
        Ok(A::a_deserialize(inctx)?.into())
    }

    fn rextend(&self, mentionable: Self::Mtbl, tail: &[u8]) -> ParseResult<'a, Ctx, Self> {
        Ok(A::a_extend(mentionable.atomic, tail)?.into())
    }
}

/// Extension trait to provide method-like utilities associated with [AtomicObject]s.
pub trait AtomicObjectExt: AtomicBase {
    /// Shorthand for getting specific [`AtomicFactory`].
    fn f() -> AtomicFactory<Self> {
        AtomicFactory::new()
    }
    /// Shorthand for getting specific [`AtomicObject`].
    fn m(self) -> AtomicObject<Self> {
        self.into()
    }
}

impl<A: AtomicBase> AtomicObjectExt for A {}
